# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import mock

from paasta_tools import adhoc_tools
from paasta_tools.utils import DeploymentsJsonV2
from paasta_tools.utils import NoConfigurationForServiceError


def test_get_default_interactive_config():
    with mock.patch(
        "paasta_tools.adhoc_tools.load_adhoc_job_config", autospec=True
    ) as mock_load_adhoc_job_config:
        mock_load_adhoc_job_config.return_value = adhoc_tools.AdhocJobConfig(
            service="fake_service",
            instance="interactive",
            cluster="fake_cluster",
            config_dict={},
            branch_dict={"deploy_group": "fake_deploy_group"},
        )
        result = adhoc_tools.get_default_interactive_config(
            "fake_service", "fake_cluster", "/fake/soa/dir", load_deployments=False
        )
        assert result.get_cpus() == 4
        assert result.get_mem() == 10240
        assert result.get_disk() == 1024


def test_get_default_interactive_config_reads_from_tty():
    with mock.patch(
        "paasta_tools.adhoc_tools.prompt_pick_one", autospec=True
    ) as mock_prompt_pick_one, mock.patch(
        "paasta_tools.adhoc_tools.load_adhoc_job_config", autospec=True
    ) as mock_load_adhoc_job_config, mock.patch(
        "paasta_tools.adhoc_tools.load_v2_deployments_json", autospec=True
    ) as mock_load_deployments_json:
        mock_prompt_pick_one.return_value = "fake_deploygroup"
        mock_load_adhoc_job_config.side_effect = NoConfigurationForServiceError
        mock_load_deployments_json.return_value = DeploymentsJsonV2(
            service="fake-service",
            config_dict={
                "deployments": {
                    "fake_deploygroup": {
                        "docker_image": mock.sentinel.docker_image,
                        "git_sha": mock.sentinel.git_sha,
                    }
                },
                "controls": {},
            },
        )
        result = adhoc_tools.get_default_interactive_config(
            "fake_service", "fake_cluster", "/fake/soa/dir", load_deployments=True
        )
        assert result.get_deploy_group() == "fake_deploygroup"
        assert result.get_docker_image() == mock.sentinel.docker_image


def test_adhoc_config_node_selectors_in_pod_spec():
    """Test that node_selectors from adhoc config appear in the generated pod spec."""
    with mock.patch(
        "paasta_tools.kubernetes_tools.load_system_paasta_config", autospec=True
    ) as mock_load_system_config, mock.patch(
        "paasta_tools.utils.load_system_paasta_config", autospec=True
    ) as mock_load_system_config_utils:
        mock_load_system_config.return_value.get_cluster_aliases.return_value = []
        mock_load_system_config.return_value.get_volumes.return_value = []
        mock_load_system_config.return_value.get_dockercfg_location.return_value = (
            "file:///root/.dockercfg"
        )
        # Copy the mock to utils as well
        mock_load_system_config_utils.return_value = (
            mock_load_system_config.return_value
        )

        # Create an adhoc config with node_selectors
        adhoc_config = adhoc_tools.AdhocJobConfig(
            service="test_service",
            instance="test_instance",
            cluster="test_cluster",
            config_dict={
                "cpus": 1.0,
                "mem": 1024,
                "disk": 1024,
                "node_selectors": {
                    "instance_type": "c5.xlarge",
                    "custom_label": [
                        {"operator": "In", "values": ["value1", "value2"]}
                    ],
                },
            },
            branch_dict={
                "docker_image": "docker-dev.yelpcorp.com/test:latest",
                "git_sha": "abc123",
                "desired_state": "start",
                "force_bounce": None,
            },
        )

        # Wrap in EksDeploymentConfig like remote_run does
        from paasta_tools.eks_tools import EksDeploymentConfig

        eks_config = EksDeploymentConfig(
            service=adhoc_config.service,
            cluster=adhoc_config.cluster,
            instance=adhoc_config.instance,
            config_dict=adhoc_config.config_dict,
            branch_dict=adhoc_config.branch_dict,
            soa_dir="/fake/soa/dir",
        )

        # Generate the pod template spec
        pod_template_spec = eks_config.get_pod_template_spec(
            git_sha="abc123",
            system_paasta_config=mock_load_system_config.return_value,
        )

        # Verify node selectors are in the pod spec
        # instance_type gets translated to canonical kubernetes label
        assert (
            pod_template_spec.spec.node_selector["node.kubernetes.io/instance-type"]
            == "c5.xlarge"
        )

        # Complex selector should be in node affinity
        assert pod_template_spec.spec.affinity is not None
        assert pod_template_spec.spec.affinity.node_affinity is not None
        node_affinity = pod_template_spec.spec.affinity.node_affinity
        assert (
            node_affinity.required_during_scheduling_ignored_during_execution
            is not None
        )
        node_selector_terms = (
            node_affinity.required_during_scheduling_ignored_during_execution.node_selector_terms
        )
        assert len(node_selector_terms) > 0

        # Find the custom_label requirement
        all_requirements = []
        for term in node_selector_terms:
            all_requirements.extend(term.match_expressions)

        custom_label_requirements = [
            req for req in all_requirements if req.key == "custom_label"
        ]
        assert len(custom_label_requirements) > 0
        req = custom_label_requirements[0]
        assert req.operator == "In"
        assert set(req.values) == {"value1", "value2"}



import os
import subprocess
import sys
from subprocess import PIPE


def cmd(args, capture_output=True):
    try:
        return subprocess.run(
            args.split(" "),
            stdout=PIPE if capture_output else None,
            stderr=PIPE if capture_output else None,
            check=True,
            text=True,
        )
    except subprocess.CalledProcessError as exc:
        if exc.stdout:
            print(exc.stdout)
        if exc.stderr:
            print(exc.stderr, file=sys.stderr)
        raise


def start_paasta_api():
    print("Starting Paasta API Server")
    p = subprocess.Popen(
        "python -m paasta_tools.api.api -D -c {} {}".format(
            os.environ["KIND_CLUSTER"], os.environ["PAASTA_API_PORT"]
        ).split(" ")
    )
    return p


def paasta_apply():
    print("Applying SOA configurations")
    service_instances = cmd("python -m paasta_tools.list_kubernetes_service_instances")
    cmd(
        "python -m paasta_tools.setup_kubernetes_job {} -v".format(
            service_instances.stdout.strip()
        ),
        True,
    )


def init_all():
    paasta_apply()



# Copyright 2015-2016 Yelp Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import multiprocessing
import os
import sys
import time
from functools import lru_cache
from typing import Optional

from paasta_tools.utils import atomic_file_write
from paasta_tools.utils import load_system_paasta_config

try:
    from okta_auth import get_and_cache_jwt_default
    from vault_tools.oidc import get_instance_oidc_identity_token
except ImportError:

    def get_instance_oidc_identity_token(
        role: str, ecosystem: Optional[str] = None
    ) -> str:
        return ""

    def get_and_cache_jwt_default(
        client_id: str, refreshable: bool = False, force: bool = False
    ) -> str:
        return ""


MAX_SSO_REFRESH_ATTEMPTS = 3


@lru_cache(maxsize=1)
def get_service_auth_token() -> str:
    """Uses instance profile to authenticate with Vault and generate token for service authentication"""
    vault_role = load_system_paasta_config().get_service_auth_vault_role()
    return get_instance_oidc_identity_token(vault_role)


def get_sso_auth_token(paasta_apis: bool = False) -> str:
    """Generate an authentication token for the calling user from the Single Sign On provider

    :param bool paasta_apis: authenticate for PaaSTA APIs
    """
    system_config = load_system_paasta_config()
    client_id = (
        system_config.get_api_auth_sso_oidc_client_id()
        if paasta_apis
        else system_config.get_service_auth_sso_oidc_client_id()
    )
    return get_and_cache_jwt_default(client_id)


def maintain_valid_sso_token(
    directory: str,
    token_file: str = "token",
    check_interval: int = 300,
    max_token_age: int = 2400,
) -> None:
    """Generates and keeps up to date SSO token in a given directory.
    The resulting token path is "{directory}/{token_file}".
    The function starts a process that never returns unless the parent process terminates.

    :param str directory: where to place the token file
    :param str token_file: name of the token file
    :param int check_interval: how often to check if the file needs to be updated (in seconds)
    :param int max_token_age: how long to wait before refreshing the token file (in seconds)
    """
    token_path = os.path.join(directory, token_file)
    system_config = load_system_paasta_config()
    client_id = system_config.get_service_auth_sso_oidc_client_id()

    # We do a first call synchronously to ensure the user has control on the input;
    # force=True ensures that we get a refresh token in memory.
    os.makedirs(directory, exist_ok=True)
    token_val = get_and_cache_jwt_default(client_id, refreshable=True, force=True)
    with atomic_file_write(token_path, mode=0o0600) as f:
        f.write(token_val)

    def auth_handler() -> None:
        while os.getppid() != 1:  # stop when parent process terminates
            time.sleep(check_interval)
            if time.time() - os.path.getmtime(token_path) < max_token_age:
                continue
            for i in range(MAX_SSO_REFRESH_ATTEMPTS):
                try:
                    # we force=True again because the library has validation tolerances of a few seconds,
                    # while we aim at monitoring the file only every few minutes
                    token_val = get_and_cache_jwt_default(
                        client_id, refreshable=True, force=True
                    )
                    break
                except Exception as e:
                    if i == MAX_SSO_REFRESH_ATTEMPTS - 1:
                        raise e
                    print(f"Error refreshing SSO token: {e}", file=sys.stderr)
                    time.sleep(5)
            with atomic_file_write(token_path, mode=0o0600) as f:
                f.write(token_val)
        os.remove(token_path)

    proc = multiprocessing.Process(target=auth_handler)
    proc.start()
